@@ -1,6 +1,8 @@
 #!perl -w
-
 use strict;
+use FindBin qw($Bin);
+use lib "$Bin/../lib";
+
 use Acme::Perl::VM;
 
 sub f{
@@ -13,8 +15,10 @@ run_block {
 	local $| = 1;
 
 	my $sum = 0;
-	for(my $i = 1; $i <= 1000; $i++){
-		$sum += f($i);
+	for(my $i = 1; $i <= 100; $i++){
+		for(my $j = 1; $j <= 10; $j++){
+			$sum += f($i * $j);
+		}
 	}
 
 	print "\n", $sum, "\n";
