@@ -1,6 +1,8 @@
 #!perl -w
-
 use strict;
+use FindBin qw($Bin);
+use lib "$Bin/../lib";
+
 use Acme::Perl::VM;
 
 sub f{
