@@ -5,7 +5,7 @@ use strict;
 use warnings;
 
 BEGIN{
-	require version; our $VERSION = version::qv('0.0.2');
+	require version; our $VERSION = version::qv('0.0.3');
 }
 
 use constant APVM_DEBUG  => ($ENV{APVM_DEBUG} || do{ our $VERSION->is_alpha || 0 });
@@ -13,6 +13,7 @@ use constant {
 	APVM_TRACE => scalar(APVM_DEBUG =~ /\b trace \b/xmsi),
 	APVM_SCOPE => scalar(APVM_DEBUG =~ /\b scope \b/xmsi),
 	APVM_CX    => scalar(APVM_DEBUG =~ /\b (?: cx | context ) \b/xmsi),
+	APVM_STACK => scalar(APVM_DEBUG =~ /\b stack \b/xmsi),
 
 	APVM_DUMMY => scalar(APVM_DEBUG =~ /\b dummy \b/xmsi),
 };
@@ -36,6 +37,7 @@ BEGIN{
 		GET_TARGET
 		GET_TARGETSTACKED
 		GET_ATARGET
+		MAXARG
 
 		PUSHBLOCK POPBLOCK TOPBLOCK
 		PUSHSUB POPSUB
@@ -60,10 +62,12 @@ BEGIN{
 
 		GVOP_gv
 
-		hv_scalar
 		vivify_ref
 		sv_newmortal sv_mortalcopy sv_2mortal
 		SvPV SvNV SvIV SvTRUE
+		av_assign av_store
+		hv_store hv_store_ent hv_scalar
+
 		defoutgv
 		gv_fullname
 
@@ -71,16 +75,31 @@ BEGIN{
 		sv_defined is_null is_not_null
 		mark_list
 		not_implemented
-		dump_object dump_value dump_stacks
+		dump_object dump_value dump_stack dump_si
+
+		apvm_extern
+		cv_external
+
 		APVM_DEBUG APVM_DUMMY
 		APVM_SCOPE APVM_TRACE
 	);
 	our %EXPORT_TAGS = (
 		perl_h => \@EXPORT_OK,
 	);
+
+	
+
+	if(-t *STDERR){
+		require Term::ANSIColor;
+
+		*deb = \&_deb_colored;
+	}
+	else{
+		*deb = \&_deb;
+	}
 }
 
-use Scalar::Util qw(looks_like_number);
+use Scalar::Util qw(looks_like_number refaddr);
 use Carp ();
 
 use Acme::Perl::VM::Context;
@@ -116,17 +135,24 @@ our %PL_ppaddr;
 	}
 }
 
-if(APVM_TRACE){
-	$PL_runops = \&runops_trace;
+if(APVM_TRACE || APVM_STACK){
+	$PL_runops = \&runops_debug;
 }
 
+our $color = 'GREEN BOLD';
+
+sub not_implemented;
+
 sub runops_standard{ # run.c
 	1 while(${ $PL_op = &{$PL_ppaddr{ $PL_op->name } || not_implemented($PL_op->ppaddr)} });
 	return;
 }
 
-sub _trace{
-	deb '.%s', uc($PL_op->name);
+sub _op_trace{
+	my $flags = $PL_op->flags;
+	my $name  = $PL_op->name;
+
+	deb '.%s', $name;
 	if(ref($PL_op) eq 'B::COP'){
 		deb '(%s%s %s:%d)', 
 			($PL_op->label ? $PL_op->label.': ' : ''),
@@ -134,56 +160,90 @@ sub _trace{
 			$PL_op->file, $PL_op->line,
 		;
 	}
-	elsif($PL_op->name eq 'entersub'){
+	elsif($name eq 'entersub'){
 		my $gv = TOP;
 		if(!$gv->isa('B::GV')){
 			$gv = $gv->GV;
 		}
-		deb '[%s]', gv_fullname($gv, '&');
+		deb '(%s)', gv_fullname($gv, '&');
 	}
-	elsif($PL_op->name eq 'aelemfast'){
+	elsif($name eq 'aelemfast'){
 		my $name;
-		if($PL_op->flags & OPf_SPECIAL){
+		if($flags & OPf_SPECIAL){
 			my $padname = $PL_comppad_name->ARRAYelt($PL_op->targ);
-			$name = $padname->POK ? '@'.$padname->PV : '[...]';
+			$name = $padname->POK ? '@'.$padname->PVX : '[...]';
 		}
 		else{
 			$name = gv_fullname(GVOP_gv($PL_op), '@');
 		}
 		deb '[%s[%s]]', $name, $PL_op->private;
-
 	}
-	elsif($PL_op->targ && $PL_op->name !~ /leave/){
-		my $padname = $PL_comppad_name->ARRAYelt($PL_op->targ);
-		if($padname->POK){
-			deb '[%s]', $padname->PVX;
-			deb ' INTRO' if $PL_op->private & OPpLVAL_INTRO;
-		}
-		elsif($PL_op->name eq 'const'){
+	elsif($PL_op->targ && $name !~ /leave/){
+		if($name eq 'const' || $name eq 'method_named'){
 			my $sv = PAD_SV($PL_op->targ);
 
-			deb '[%s]', $sv->POK ? B::perlstring($sv->PVX) : $sv->as_string;
+			if(is_scalar($sv)){
+				deb '(%s)', $sv->POK ? B::perlstring($sv->PVX) : $sv->as_string;
+			}
+			else{
+				deb '(%s)', ddx([$sv->object_2svref])->Indent(0)->Dump;
+			}
 		}
+		else{
+			my $padname = $PL_comppad_name->ARRAYelt($PL_op->targ);
+			if($padname->POK){
+				deb '(%s)', $padname->PVX;
+				deb ' INTRO' if $PL_op->private & OPpLVAL_INTRO;
+			}
+		}
+	}
+	elsif($PL_op->can('sv')){
+		my $sv = SVOP_sv($PL_op);
+		if($sv->class eq 'GV'){
+			my $prefix = $name eq 'gvsv' ? '$' : '*';
+			deb '(%s)', gv_fullname($sv, $prefix);
+			deb ' INTRO' if $PL_op->private & OPpLVAL_INTRO;
+		}
+		else{
+			deb '(%s)', B::perlstring(SvPV(SVOP_sv($PL_op)));
+		}
+	}
 
-	}
-	elsif($PL_op->can('gv')){ # GVOP or PADOPgv->object_2svref
-		my $prefix = $PL_op->name eq 'gvsv' ? '$' : '*';
-		my $gv = GVOP_gv($PL_op);
-		deb '[%s]', gv_fullname(GVOP_gv($PL_op), $prefix);
-	}
+	deb ' VOID'    if( ($flags & OPf_WANT) == OPf_WANT_VOID   );
+	deb ' SCALAR'  if( ($flags & OPf_WANT) == OPf_WANT_SCALAR );
+	deb ' LIST'    if( ($flags & OPf_WANT) == OPf_WANT_LIST   );
+
+	deb ' KIDS'    if $flags & OPf_KIDS;
+	deb ' PARENS'  if $flags & OPf_PARENS;
+	deb ' REF'     if $flags & OPf_REF;
+	deb ' MOD'     if $flags & OPf_MOD;
+	deb ' STACKED' if $flags & OPf_STACKED;
+	deb ' SPECIAL' if $flags & OPf_SPECIAL;
 
 	deb "\n";
 }
 
-sub runops_trace{
-	_trace();
+sub runops_debug{
+	_op_trace();
 	while(${ $PL_op = &{$PL_ppaddr{ $PL_op->name } || not_implemented($PL_op->ppaddr)} }){
-		_trace();
+		if(APVM_STACK){
+			dump_stack();
+		}
+
+		_op_trace();
+	}
+	if(APVM_STACK){
+		dump_stack();
 	}
 	return;
 }
 
-sub deb{
+sub _deb_colored{
+	my($fmt, @args) = @_;
+	printf STDERR Term::ANSIColor::colored($fmt, $color), @args;
+	return;
+}
+sub _deb{
 	my($fmt, @args) = @_;
 	printf STDERR $fmt, @args;
 	return;
@@ -203,8 +263,6 @@ sub longmess{
 		my $cx   = $PL_cxstack[$cxix];
 		my $cop  = $cx->oldcop;
 
-		last if is_null($cop);
-
 		my $args;
 
 		if($cx->argarray){
@@ -287,24 +345,25 @@ sub GET_ATARGET{
 	return $PL_op->flags & OPf_STACKED ? $PL_stack[$#PL_stack-1] : PAD_SV($PL_op->targ);
 }
 
+sub MAXARG{
+	return $PL_op->private & 0x0F;
+}
+
 sub PUSHBLOCK{
-	my($type, $sp, $gimme) = @_;
+	my($type, %args) = @_;
 
-	my $cx_class = 'Acme::Perl::VM::Context::' . $type;
+	$args{oldcop}     = $PL_curcop;
+	$args{oldmarksp}  = $#PL_markstack;
+	$args{oldscopesp} = $#PL_scopestack;
 
-	push @PL_cxstack, $cx_class->new(
-		oldsp      => $sp,
-		oldcop     => $PL_curcop,
-		oldmarksp  => $#PL_markstack,
-		oldscopesp => $#PL_scopestack,
-		gimme      => $gimme,
-	);
+	my $cx = "Acme::Perl::VM::Context::$type"->new(\%args);
+	push @PL_cxstack, $cx;
 
 	if(APVM_CX){
 		deb "%s" . "Entering %s\n", (q{>} x @PL_cxstack), $type;
 	}
 
-	return;
+	return $cx;
 }
 
 sub POPBLOCK{
@@ -330,13 +389,6 @@ sub TOPBLOCK{
 	return $cx;
 }
 
-sub PUSHSUB{
-	my($cx, %args) = @_;
-	$cx->cv($args{cv});
-	$cx->hasargs($args{hasargs});
-	$cx->olddepth($args{cv}->DEPTH);
-	return;
-}
 sub POPSUB{
 	my($cx) = @_;
 	if($cx->hasargs){
@@ -346,23 +398,6 @@ sub POPSUB{
 	}
 	return;
 }
-sub PUSHLOOP{
-	my($cx, %args) = @_;
-
-	$cx->resetsp($args{sp});
-
-	$cx->label($PL_curcop->label);
-	$cx->myop($PL_op);
-	$cx->nextop($PL_op->nextop);
-
-	if($args{iterdata}){ # foreach
-		$cx->padvar($args{padvar});
-		$cx->for_def($args{for_def});
-
-		$cx->ITERDATA_SET($args{iterdata});
-	}
-	return;
-}
 
 sub POPLOOP{
 	my($cx) = @_;
@@ -414,7 +449,7 @@ sub ENTER{
 
 sub LEAVE{
 	my $oldsave = pop @PL_scopestack;
-	LEAVE_SCOPE($oldsave) if $oldsave < $#PL_savestack;
+	LEAVE_SCOPE($oldsave);
 
 	if(APVM_SCOPE){
 		deb "%s" . "LEAVE\n", ('>' x (@PL_scopestack+1));
@@ -428,8 +463,8 @@ sub LEAVE_SCOPE{
 		my $ss = pop @PL_savestack;
 
 		if(APVM_SCOPE){
-			deb "%s" . "leave %s saved %s\n",
-				('>' x (@PL_cxstack+1)), $ss->type, $ss->saved_at;
+			deb "%s" . "leave %s %s\n",
+				('>' x (@PL_cxstack+1)), $ss->type, $ss->saved_state;
 		}
 		$ss->leave();
 	}
@@ -532,13 +567,13 @@ sub dopoptosub{
 	return -1;
 }
 
+my %loop;
+@loop{qw(SUBST SUB EVAL NULL)} = ();
+$loop{LOOP}    = TRUE;
+
 sub dopoptoloop{
 	my($startingblock) = @_;
 
-	my %loop;
-	@loop{qw(SUBST SUB EVAL NULL)} = ();
-	$loop{LOOP}    = TRUE;
-
 	for(my $i = $startingblock; $i >= 0; --$i){
 		my $cx   = $PL_cxstack[$i];
 		my $type = $cx->type;
@@ -555,9 +590,6 @@ sub dopoptoloop{
 }
 sub dopoptolabel{
 	my($label) = @_;
-	my %loop;
-	@loop{qw(SUBST SUB EVAL NULL)} = ();
-	$loop{LOOP}    = TRUE;
 
 	for(my $i = $#PL_cxstack; $i >= 0; --$i){
 		my $cx   = $PL_cxstack[$i];
@@ -634,19 +666,15 @@ sub LVRET(){ # cf. is_lvalue_sub() in pp_ctl.h
 	return FALSE;
 }
 
+sub SVOP_sv{
+	my($op) = @_;
+	return USE_ITHREADS ? PAD_SV($op->padix) : $op->sv;
+}
 sub GVOP_gv{
 	my($op) = @_;
-
 	return USE_ITHREADS ? PAD_SV($op->padix) : $op->gv;
 }
 
-sub hv_scalar{
-	my($hv) = @_;
-	my $sv = sv_newmortal();
-	$sv->setval(scalar %{ $hv->object_2svref });
-	return $sv;
-}
-
 sub vivify_ref{
 	not_implemented 'vivify_ref';
 }
@@ -720,6 +748,44 @@ sub SvIV{
 	return int(${$ref});
 }
 
+sub av_assign{
+	my $av   = shift;
+	my $ref  = $av->object_2svref;
+	$#{$ref} = $#_;
+	for(my $i = 0; $i < @_; $i++){
+		tie $ref->[$i], 'Acme::Perl::VM::Alias', $_[$i]->object_2svref;
+	}
+	return;
+}
+
+sub av_store{
+	my($av, $ix, $sv) = @_;
+	tie $av->object_2svref->[$ix],
+		'Acme::Perl::VM::Alias', $sv->object_2svref;
+	return;
+}
+
+sub hv_store{
+	my($hv, $key, $sv) = @_;
+	tie $hv->object_2svref->{$key},
+		'Acme::Perl::VM::Alias', $sv->object_2svref;
+	return;
+}
+
+sub hv_store_ent{
+	my($hv, $key, $sv) = @_;
+	tie $hv->object_2svref->{ ${$key->object_2svref} },
+		'Acme::Perl::VM::Alias', $sv->object_2svref;
+	return;
+}
+
+sub hv_scalar{
+	my($hv) = @_;
+	my $sv = sv_newmortal();
+	$sv->setval(scalar %{ $hv->object_2svref });
+	return $sv;
+}
+
 sub defoutgv{
 	no strict 'refs';
 	return \*{ select() };
@@ -756,21 +822,59 @@ sub is_null{
 	return !${$sv};
 }
 
+my %not_a_scalar;
+@not_a_scalar{qw(AV HV CV IO)} = ();
+sub is_scalar{
+	my($sv) = @_;
+	return !exists $not_a_scalar{ $sv->class };
+}
+
 sub mark_list{
 	my($mark) = @_;
 	return map{ ${ $_->object_2svref } } splice @PL_stack, $mark+1;
 }
 
+{
+	our %external;
+
+	sub apvm_extern{
+		foreach my $arg(@_){
+			if(ref $arg){
+				if(ref($arg) ne 'CODE'){
+					Carp::croak('Not a CODE reference for apvm_extern()');
+				}
+				$external{refaddr $arg} = 1;
+			}
+			else{
+				my $stash = do{ no strict 'refs'; \%{$arg .'::'} };
+				while(my $name = each %{$stash}){
+					my $code_ref = do{ no strict 'refs'; *{$arg . '::' . $name}{CODE} };
+					if(defined $code_ref){
+						$external{refaddr $code_ref} = 1;
+					}
+				}
+			}
+		}
+		return;
+	}
+
+	sub cv_external{
+		my($cv) = @_;
+		return $cv->XSUB || $external{ ${$cv} };
+	}
+}
+
 sub ddx{
 	require Data::Dumper;
 	my $ddx = Data::Dumper->new(@_);
 	$ddx->Indent(1);
-	$ddx->Terse(1);
-	$ddx->Quotekeys(0);
+	$ddx->Terse(TRUE);
+	$ddx->Quotekeys(FALSE);
+	$ddx->Useqq(TRUE);
 	return $ddx if defined wantarray;
 
 	my $name = ( split '::', (caller 2)[3] )[-1];
-	print STDERR $name, ': ', $ddx->Dump();
+	print STDERR $name, ': ', $ddx->Dump(), "\n";
 	return;
 }
 sub dump_object{
@@ -780,22 +884,98 @@ sub dump_object{
 sub dump_value{
 	ddx([\@_]);
 }
-sub dump_stacks{
-	my $stacks = {
+
+
+sub dump_stack{
+	require Data::Dumper;
+	no warnings 'once';
+
+	local $Data::Dumper::Indent    = 0;
+	local $Data::Dumper::Terse     = TRUE;
+	local $Data::Dumper::Quotekeys = FALSE;
+	local $Data::Dumper::Useqq     = TRUE;
+
+	deb "(%s)\n", join q{,}, map{
+		# find variable name
+		my $varname = '';
+		my $class   = $_->class;
+
+		if($class eq 'SPECIAL'){
+			($varname = $_->special_name) =~ s/^\&PL_//;
+			$varname;
+		}
+		elsif($class eq 'CV'){
+			$varname = '&' . gv_fullname($_->GV);
+		}
+		else{
+			for(my $padix = 0; $padix < @PL_curpad; $padix++){
+				my $padname;
+				if(${ $PL_curpad[$padix] } == ${ $_ }){
+					$padname = $PL_comppad_name->ARRAYelt($padix);
+				}
+				elsif($_->ROK && ${$PL_curpad[$padix]} == ${ $_->RV }){
+					$padname = $PL_comppad_name->ARRAYelt($padix);
+					$varname .= '\\';
+				}
+
+				if($padname){
+					if($padname->POK){
+						$varname .= $padname->PVX . ' ';
+					}
+					last;
+				}
+			}
+			$varname . Data::Dumper->Dump([is_scalar($_) ? ${$_->object_2svref} : $_->object_2svref], [$_->ROK ? 'SV' : '*SV']);
+		}
+
+	} @PL_stack;
+
+	return;
+}
+sub _dump_stack{
+	my $warn;
+	my $ddx = ddx([[map{
+			if(ref $_){
+				is_scalar($_) ? ${$_->object_2svref} : $_->object_2svref;
+			}
+			else{
+				$warn++;
+				$_;
+			}
+	} @PL_stack]], ['*PL_stack']);
+	$ddx->Indent(0);
+	deb "  %s\n", $ddx->Dump();
+
+	if($warn){
+		apvm_die 'No sv found (%d)', $warn;
+	}
+	return;
+}
+
+sub dump_si{
+	my %stack_info = (
 		stack     => \@PL_stack,
 		markstack => \@PL_markstack,
 		cxstack   => \@PL_cxstack,
 		scopstack => \@PL_scopestack,
 		savestack => \@PL_savestack,
 		tmps      => \@PL_tmps,
-	};
+	);
 
-	ddx([$stacks]);
+	ddx([\%stack_info]);
 }
 
 sub not_implemented{
-	my $opname = $PL_op && is_not_null($PL_op) ? $PL_op->name . ': ' : '';
-	@_ =sprintf('%s%s is not implemented', $opname, join('', @_));
+	if(!@_){
+		if($PL_op && is_not_null($PL_op)){
+			@_ = ($PL_op->name);
+		}
+		else{
+			@_ = (caller 0)[3];
+		}
+	}
+
+	push @_, ' is not implemented';
 	goto &Carp::confess;
 }
 
@@ -810,16 +990,17 @@ sub call_sv{ # perl.h
 
 	my $cv = $sv->toCV();
 
-	$PL_curcop ||= bless \do{ my $addr = 0 }, 'B::COP'; # dummy cop
+	my $old_op  = $PL_op;
+	my $old_cop = $PL_curcop;
 
-	my $oldop = $PL_op;
 	$PL_op = Acme::Perl::VM::OP_CallSV->new(
 		cv    => $cv,
 		next  => NULL,
 		flags => OP_GIMME_REVERSE($flags),
 	);
-	PUSH($cv);
+	$PL_curcop = $PL_op;
 
+	PUSH($cv);
 	my $oldmark  = TOPMARK;
 
 	$PL_runops->();
@@ -833,7 +1014,8 @@ sub call_sv{ # perl.h
 		LEAVE;
 	}
 
-	$PL_op = $oldop;
+	$PL_op     = $old_op;
+	$PL_curcop = $old_cop;
 
 	return $retval;
 }
@@ -872,7 +1054,7 @@ sub run_block(&@){
 
 package
 	Acme::Perl::VM::OP_CallSV;
-	
+
 use Mouse;
 
 has cv => (
@@ -896,11 +1078,37 @@ has flags => (
 	required => 1,
 );
 
+sub class(){ 'OP' }
 sub name(){ 'entersub' }
 sub desc(){ 'subroutine entry' }
 
+sub file(){ __FILE__ }
+sub line(){ 0 }
+
+sub isa{
+	shift;
+	return B::COP->isa(@_);
+}
+
 __PACKAGE__->meta->make_immutable();
 
+package
+	Acme::Perl::VM::Alias;
+
+
+sub TIESCALAR{
+	my($class, $scalar_ref) = @_;
+	return bless [$scalar_ref], $class;
+}
+sub FETCH{
+	return ${ $_[0]->[0] }
+}
+sub STORE{
+	${ $_[0]->[0] } = $_[1];
+	return;
+}
+
+
 1;
 __END__
 
@@ -944,7 +1152,7 @@ F<pp.h> for PUSH/POP macros.
 
 F<pp.c>, F<pp_ctl.c>, and F<pp_hot.c> for ppcodes.
 
-C<op.h> for opcodes.
+F<op.h> for opcodes.
 
 F<cop.h> for COP and context blocks.
 
